#!perl -Tw
use strict;
use CGI::Carp 'fatalsToBrowser';
require 5.004;

use vars qw($VERSION $mailHost $webmaster $failFile $default_is_required
            $serverName $base_url $debug $cgi_save);

$VERSION = sprintf("%d.%02d", q$Revision: 1.0 $ =~ /(\d+)\.(\d+)/);

######################   Configuration   ######################
##############  See POD at end for documentation  #############

# default SMTP host to send mail to
$mailHost = 'my.mail.host';

# mail address of this site's webmaster
$webmaster = 'webmaster@my.domain';

# filename to save the mail to if sending it fails
$failFile = '/any/path/formmail.mails';

# default value for is_required
$default_is_required = 0;

# debugging options; remove comment when debugging
#
# contents is filename to redirect debug output (via STDERR) to
#$debug = '/any/path/formmail_debug.log';
#
# remove comment when debugging; contents is CGI::save filename
#$cgi_save = '/any/path/formmail.cgi_save';

####################   End Configuration   ####################
########  No user serviceable parts beyond this point  ########

use CGI ':html';
use URI::URL;
use LWP::UserAgent;
use HTTP::Request;

# re-open STDERR if debugging
if (defined $debug)
{
	open STDERR, ">>$debug" or die "Can't open $debug: $!";
	printf STDERR '--- Start processing --- '.localtime()."\n";
}
else
{
	$debug = 0;
}

###############################################################

#
# Define class FormMail as derived from HTML::Parser
#
{
	package FormMail;

	use vars qw(@ISA $AUTOLOAD);
	use HTML::Parser;

	@ISA = 'HTML::Parser';

	sub new ($$$)
	{
		my ($class, $myurl, $formurl) = @_;
		my $self = $class->SUPER::new;

		# These fields must be named like
		# prefix 'FM' + the respective command
		$self->{FMsender}       = '';
		$self->{FMsubject}      = '';
		$self->{FMsuccess_url}  = '';
		$self->{FMfailure_url}  = '';
		$self->{FMprocess}      = '';
		$self->{FMpre_results}  = '';
		$self->{FMpost_results} = '';

		$self->{FMrecipient}    = '';
		$self->{FMmy_url}       = URI::URL::url( $myurl )->path;
		$self->{FMform_url}     = URI::URL::url( $formurl )->path;

		$self->{FMelements}     = [];
		$self->{FMfields}       = [];
		$self->{FMvalues}       = \0;

		return bless $self, $class;
	}

	sub AUTOLOAD ($)
	{
		my ($self) = @_;

		return if $AUTOLOAD =~ /::DESTROY$/;
		$AUTOLOAD =~ s/^FormMail::/FM/;

		die "No such variable '$AUTOLOAD'"
			if !exists $self->{$AUTOLOAD};

		return $self->{$AUTOLOAD};
	}

	sub set_values ($$)
	{
		my ($self, $values) = @_;
		$self->{FMvalues} = $values;
	}

	sub _eval_expression ($$)
	{
		my ($expression, $value) = @_;

		# very simple for now ...
		my ($var, $op, $val) = split /\s+/, $expression, 3;
		my $result = eval "defined \$value->{\$var} && (\$value->{\$var} $op '\$val')";
		die "Error in expression \"$expression\": $@\n" if $@;
		return $result;
	}

	#
	# HTML::Parser called methods
	#

	sub declaration ($$)
	{
		my ($self, $decl) = @_;
		push @{$self->{FMelements}}, ['DECL', $decl];
	}

	sub start ($$$$$)
	{
		my ($self, $tag, $attr) = @_;

		if ($tag eq 'form')
		{
			$attr->{action} =~ /^mailto:(\S+)/i
				or die "FormMail form requires a 'mailto' action";
			
			$self->{FMrecipient} = $1;

			# rewrite method to POST
			$attr->{method} = 'POST';

			# rewrite action to point to ourself
			$attr->{action} = $self->{FMmy_url};

			push @{$self->{FMelements}}, [$tag, $attr];

			# Insert hidden FormUrl field
			push @{$self->{FMelements}},
				[
					'INPUT', # use uppercase to differentiate from normal input fields
					{
						type  => 'hidden',
						name  => 'FormURL',
						value => $self->{FMform_url}
					}
				];
		}
		else
		{
			push @{$self->{FMelements}}, [$tag, $attr];
			if (  (($tag eq 'input' && $attr->{type} !~ /^(?:submit|reset)$/i)
			       || $tag =~ /^(?:textarea|select)$/)
			    && defined $attr->{name})
			{
				push @{$self->{FMfields}}, $attr->{name};
			}
		}
	}

	sub end ($$)
	{
		my ($self, $tag) = @_;

		if ($tag eq 'head')
		{
			# insert webmaster's mail address
			push @{$self->{FMelements}},
				[
					'link',
					{
						'ref' => 'made',
						href  => "mailto:$::webmaster"
					}
				];
		}

		push @{$self->{FMelements}}, ['END', $tag];
	}

	sub text ($$)
	{
		my ($self, $text) = @_;
		push @{$self->{FMelements}}, ['TEXT', $text];
	}

	sub comment ($$)
	{
		my ($self, $comment) = @_;

		my ($command, $parameter) = ($comment =~ /^!(\w+)!(.*)$/s);
		if ($command)
		{
			return if $self->{FMonly_parsing};
			# comment looks like a FormMail command
			$command = lc $command;
			if ($command =~
				/^(?:sender|
				     subject|
				     success_url|
				     failure_url|
				     process|
				     pre_results|
				     post_results)
				$/x)
			{
				for ($parameter) { s/^\s+//; s/\s+$//; }
				$parameter ne ''
					or die "FormMail command '$command' requires parameter";
				$self->{"FM$command"} = $parameter;
			}
			elsif ($command =~ /^(?:not_)?required$/)
			{
				push @{$self->{FMelements}}, ['CMD', $command];
			}
			elsif ($command =~ /^(?:not_)?required_if$/)
			{
				for ($parameter) { s/^\s+//; s/\s+$//; }
				$parameter ne ''
					or die "FormMail command '$command' requires parameter";
				push @{$self->{FMelements}}, ['CMD', $command, $parameter];
			}
			elsif ($command =~
				/^(?:only_if_missing_required|
				     only_if_first_time|
				     only_if_process_fails)
				$/x)
			{
				push @{$self->{FMelements}}, ['CMD', $command, $parameter];
			}
			else
			{
				die "Error: '$command' is not a valid FormMail command";
			}
		}
		else
		{
			push @{$self->{FMelements}}, ['CMNT', $comment];
		}
	}

	# End HTML::Parser called methods
	#################################

	#
	# return parsed page as HTML
	#
	# optionally takes two boolean values which are interpreted as
	# 'required values were filled out' and 'processing succeeded',
	# respectively.
	# If these are not passed, assumes the form
	# is to be output for the first time
	#
	sub as_HTML ($;$$)
	{
		my ($self, $required_ok, $process_ok) = @_;
		my $is_first_time = !defined $required_ok || !defined $process_ok;
		$required_ok = 1 if !defined $required_ok;
		$process_ok  = 1 if !defined $process_ok;

		my $value = undef;
		$value = $self->{FMvalues} if !$is_first_time;

		my $HTML = '';
		my $selectName = '';
		my $textareaName = '';

		# as this procedure relies on an empty key for recognizing
		# select and textarea sections...
		delete $value->{''} if exists $value->{''};

		ELEMENT:
		for my $element (@{$self->{FMelements}})
		{
			my ($tag, $attr) = @$element;

			($tag eq 'CMD') && do
			{
				if (   ($attr eq 'only_if_first_time'       && $is_first_time)
				    || ($attr eq 'only_if_missing_required' && !$required_ok )
				    || ($attr eq 'only_if_process_fails'    && $required_ok && !$process_ok  ))
				{
					$HTML .= $element->[2];  # Insert CMD parameter as TEXT
				}

				next ELEMENT;
			};

			($tag eq 'DECL') && do { $HTML .= "<!$attr>";     next ELEMENT; };
			($tag eq 'CMNT') && do { $HTML .= "<!--$attr-->"; next ELEMENT; };

			($tag eq 'TEXT') && do
			{
				# supress if text is a textarea default value,
				# and the user has entered something into that textarea
				$HTML .= $attr
					unless $value->{$textareaName};

				next ELEMENT;
			};

			($tag eq 'END') && do
			{
				$HTML .= "</$attr>";

				# reset select/textarea special mode
				$selectName   = '' if $attr eq 'select';
				$textareaName = '' if $attr eq 'textarea';

				next ELEMENT;
			};

			# Default case:
			#   $tag is a HTML start tag,
			#   $attr is a hash reference of attributes

			my $attrname = $attr->{name}; # the most important attribute

			# fill in user's response
			if (   $tag eq 'input'
			    && defined $attrname
			    && defined $value->{$attrname})
			{
				if ($attr->{type} =~ /^text$/i)
				{
					$attr->{value} = $value->{$attrname};
				}
				elsif ($attr->{type} =~ /^(?:radio|checkbox)$/i)
				{
					# toggle the CHECKED attribute
					if ($value->{$attrname} =~ /^$attr->{value}$/m)
					{
						$attr->{checked} = undef;
					}
					else
					{
						delete $attr->{checked};
					}
				}
			}
			elsif ($tag eq 'option' && defined $value->{$selectName})
			{
				# toggle the SELECTED attribute
				if ($value->{$selectName} =~ /^$attr->{value}$/m)
				{
					$attr->{selected} = undef;
				}
				else
				{
					delete $attr->{selected};
				}
			}
			elsif ($tag eq 'select' && defined $attrname)
			{
				$selectName = $attrname; # save for 'option' case
			}
			elsif ($tag eq 'textarea' && defined $attrname)
			{
				$textareaName = $attrname; # save for default value
			}

			# output tag and attributes
			$HTML .= "<$tag";
			for my $a (keys %$attr)
			{
				$HTML .= " $a";
				$HTML .= "=\"$attr->{$a}\"" if defined $attr->{$a};
			}
			$HTML .= '>';

			# textarea is special in that its value is output as text
			if (   $tag eq 'textarea'
			    && defined $attrname
			    && defined $value->{$attrname})
			{
				$HTML .= $value->{$attrname};
			}
		}

		return $HTML;
	}

	#
	# check presence of all required input fields
	#
	sub required_ok ($)
	{
		my ($self) = @_;
		my $value = $self->{FMvalues};

		my $is_required = $::default_is_required;
		for my $element (@{$self->{FMelements}})
		{
			my ($tag, $attr, $param) = @$element;
			if ($tag eq 'CMD')
			{
				if ($attr eq 'required_if')
				{
					$is_required = _eval_expression( $param, $value );
				}
				elsif ($attr eq 'not_required_if')
				{
					$is_required = ! _eval_expression( $param, $value );
				}
				elsif ($attr eq 'required')
				{
					$is_required = 1;
				}
				elsif ($attr eq 'not_required')
				{
					$is_required = 0;
				}
			}
			elsif (   $is_required
			       && (($tag eq 'input' && $attr->{type} !~ /^(?:submit|reset)$/i)
			           || $tag =~ /^(?:textarea|select)$/)
			       && defined $attr->{name}
			       && !$value->{$attr->{name}})
			{
				return 0
			}
		}
		return 1;
	}

	#
	# Do postprocessing
	#
	sub process_ok ($)
	{
		use FindBin;

		my ($self) = @_;

		return 1 if !$self->{FMprocess};

		my $path = $FindBin::Bin;
		$path = VMS::Filespec::unixify( $path ) if $^O eq 'VMS';

		my $process = $path . '/' . $self->{FMprocess};
		return FormMail::process( \@{$self->{FMfields}}, $self->{FMvalues} )
			if -f $process && -r _ && require $process;
		return 0;
	}

	#
	# Send mail
	#
	sub send_mail ($)
	{
		use Net::SMTP;
	
		my ($self) = @_;
		my $value = $self->{FMvalues};

		print STDERR "Sending mail...\n" if $::debug;

		# check that all required variables have been filled in
		$self->{FMsender} ne '' && $self->{FMrecipient} ne ''
			or die "Required mail parameter is missing\n";

		# needed on architectures where Net::Domain::_hostname doesn't work
		my %options = ( Hello => $::serverName );
		$options{Debug} = 1 if $::debug;
	
		my @data = (
			"From: $self->{FMsender}\n",
			"To: $self->{FMrecipient}\n",
			"Subject: $self->{FMsubject}\n",
			# This MIME assumption may of course be wrong...
			"Mime-Version: 1.0\n",
			"Content-Type: text/plain; charset=iso-8859-1\n",
			"Content-Transfer-Encoding: 8bit\n",
			#
			"\n"
		);
		push @data, $self->{FMpre_results};
		my %seen;
		for my $e (@{$self->{FMfields}})
		{
			next if $seen{$e}++ || !exists $value->{$e};
			my $blanks = ' ' x length($e);
			my @val = split "\n", $value->{$e};
			push @data, sprintf "%s: %s\n", $e     , (shift @val || '');
			push @data, sprintf "%s  %s\n", $blanks,  shift @val
				while @val != 0;
		}
		push @data, $self->{FMpost_results};
	
		my $smtp = Net::SMTP->new( $::mailHost, %options );
		my $success = 0;
		if ($smtp)
		{
			$success   = $smtp->mail( $self->{FMsender} );
			$success &&= $smtp->recipient( $self->{FMrecipient} );
			$success &&= $smtp->data( \@data );
			$smtp->quit;
		}

		warn "Failed to send mail\n" if !$success;
		if (!$success && $::failFile)
		{
			if (open FAILFILE, $::failFile)
			{
				print FAILFILE '-' x 80, "\n";
				print FAILFILE join( '', @data ), "\n";
				close FAILFILE;
			}
			else
			{
				warn "Failed to save mail either: $!\n";
			}
		}

		return $success;
	}

} # End of class FormMail

###############################################################

#
# Main program
#

sub abs_url
{
  url( url( $_[0] )->path, $base_url )->abs->as_string;
}

$| = 1;

my $cgi;
if (defined $cgi_save && $ENV{'FORMMAIL_DEBUG'})
{
	open CGI_DBG, "<$cgi_save" or die "Can't open $cgi_save: $!";
	$cgi = new CGI(\*CGI_DBG);
	close CGI_DBG;
}
else
{
	$cgi = new CGI;
	if (defined $cgi_save)
	{
		open CGI_DBG, ">$cgi_save" or die "Can't open $cgi_save: $!";
		$cgi->save(\*CGI_DBG);
		close CGI_DBG;
	}
}

$serverName = $cgi->server_name;
if ($serverName =~ /^(?:\d+\.)+\d+$/)
{
	use Socket;

	my $s = gethostbyaddr inet_aton($serverName), AF_INET;
	$serverName = $s if defined $s;
}
$base_url = "http://$serverName/";

my $cgi_url  = abs_url( $cgi->url );

my $form_url = $cgi->param('FormURL') or die "No FormURL parameter passed\n";
$cgi->delete('FormURL');

my $url = abs_url( $form_url );
print STDERR "GETting $url...\n" if $debug;
my $ua = new LWP::UserAgent;
$ua->agent( "$0/$VERSION" );
my $request = HTTP::Request->new( GET => $url );
my $response = $ua->simple_request( $request );
print STDERR 'Status: ',$response->code,' ',$response->message,"\n" if $debug;

unless ($response->is_success)
{
	print STDERR "Sending GET failure message...\n" if $debug;

	print
		$cgi->header,
		start_html( 'FormMail error', $webmaster ),
		'An error occured when retrieving the form ',
		$url,
		p,
		$response->error_as_HTML,
		end_html;

	exit;
}

my $fmp = FormMail->new( $cgi_url, $form_url );
$fmp->parse( $response->content );
$fmp->eof;

my $request_method = $cgi->request_method || '';
if ($request_method eq 'GET')
{
	print STDERR "Sending the form...\n" if $debug;

	print
		$cgi->header(-Pragma=>'No-Cache'),
		$fmp->as_HTML;

	exit;
}

die "REQUEST_METHOD is neither GET nor POST"
	unless $request_method eq 'POST';

print STDERR "Processing form...\n" if $debug;

# get form variables, save them in %values
print STDERR "Input:\n" if $debug;
my %values;
foreach ($cgi->param)
{
	my @val = $cgi->param($_);
	$values{$_} = join "\n", @val;
	if (defined $debug)
	{
		my $blanks = ' ' x length($_);
		printf STDERR "  %s: %s\n", $_     , shift @val;
		printf STDERR "  %s  %s\n", $blanks, shift @val
			while @val != 0;
	}
}
$fmp->set_values( \%values );

my $required_ok = $fmp->required_ok();
my $process_ok  = $required_ok && $fmp->process_ok();

unless ($required_ok && $process_ok)
{
	if (defined $debug)
	{
		print STDERR "Re-sending the form ";
		if (!$required_ok)
		{
			print STDERR "because of input missing\n";
		}
		elsif (!$process_ok)
		{
			print STDERR "because of process failure\n";
		}
	}

	print
		$cgi->header(-Pragma=>'No-Cache'),
		$fmp->as_HTML($required_ok, $process_ok);

	exit;
}

unless ($fmp->send_mail())
{
	my $furl = $fmp->failure_url;
	if ($furl)
	{
		$furl = abs_url( $furl ) if $furl !~ m!^(?:http|ftp)://!;

		print STDERR "Redirecting to $furl...\n" if $debug;

		print $cgi->redirect( $furl );
	}
	else
	{
		print STDERR "Sending mail failure message...\n" if $debug;

		print
			$cgi->header,
			start_html( 'FormMail failure', $webmaster ),
			'Thank you for supplying all required fields.',
			p,
			'Unfortunately, an error occured when mailing the information to ',
			$fmp->recipient,
			'.',
			end_html;
	}

	exit;
}

# everything went well, tell the user

my $surl = $fmp->success_url;
if ($surl)
{
	$surl = abs_url( $surl ) if $surl !~ m!^(?:http|ftp)://!;

	print STDERR "Redirecting to $surl...\n" if $debug;

	print $cgi->redirect( $surl );
}
else
{
	print STDERR "Sending success message...\n" if $debug;

	print
		$cgi->header,
		start_html( 'FormMail success', $webmaster ),
		'Thank you for supplying all required fields.',
		p,
		'The information was forwarded to ',
		$fmp->recipient,
		'.',
		end_html;
}

__END__
###############################################################

=head1 NAME

B<FormMail.pl> - Handle HTML forms and mail the results

=head1 SYNOPSIS

  http://server/cgi-path/FormMail.pl?FormURL=relative_form_URL

=head1 REQUIREMENTS

This program requires Perl version 5.004. It uses the CGI, HTML::Parser and
Net::SMTP (part of the libnet library) modules, and parts
of the libwww-perl library.

=head1 DESCRIPTION

The B<FormMail> program acts differently on being called with the C<GET>
and the C<POST> request method.

When called using the C<GET> request method (like above), it takes as
its parameter (named C<FormURL>) a relative URL containing a HTML form,
retrieves it, mangles it (this includes redirecting the C<FORM ACTION>
to itself) and sends it to the client.

When the form is then submitted (with the C<POST> request method),
B<FormMail> checks the values entered against policies layed down in the
HTML form file, and either re-sends the form (inserting already filled
out fields with the values entered), or sends an E-Mail to the address
specified in the C<FORM ACTION> and optionally redirects the browser to
another URL.

=head1 COMMANDS

Commands to direct B<FormMail>'s processing are integrated into the HTML
form file as comments of the form C<E<lt>!--!command!parameter--E<gt>>.

=head2 Commands governing the requirement of input fields being filled out

All of the following commands remain in effect until the next such command
is read. An implicit C<not_required> command is assumed at the top of the
form (this default is configurable; see below under C<$default_is_required>).

=over 4

=item C<required>

All input fields up to the next requirement command are marked
as required, i.e. are checked to have a non-empty contents. The
C<parameter> is ignored.

=item C<not_required>

All input fields up to the next requirement command are marked
as optional, i.e. are not checked. The C<parameter> is ignored.

=item C<required_if>

The C<parameter> must have the form C<name eq value> or C<name ne value>.

All input fields up to the next requirement command are marked as
required, i.e. are checked to have a non-empty contents, B<if> the input
field C<name>'s value is (not) equal to C<value>.

=item C<not_required_if>

The C<parameter> must have the form C<name eq value> or C<name ne value>.

All input fields up to the next requirement command are marked as optional,
i.e. are not checked, B<if> the input field C<name>'s value is (not) equal
to C<value>.

=back

=head2 Commands governing the mailing process

=over 4

=item C<sender>

C<parameter> must be of the form C<user@domain.name>. It is used for the
C<From:> header field in the E-Mail.

=item C<subject>

C<parameter> is used as the C<Subject:> header field of the E-Mail.

=item C<pre_results>

The contents of C<parameter> is sent with the E-Mail, in front of the list
of form field name/value pairs.

=item C<post_results>

The contents of C<parameter> is sent with the E-Mail, behind the list of form
field name/value pairs.

=back

=head2 Miscellaneous commands

=item C<process>

The C<parameter> is the filename of a perl script (residing in C<FormMail>'s
own directory) which defines a FormMail::process routine. This routine is
called after all parameters have been verified and before the mail is sent.
The routine is called with two parameters: An array reference containing the
field names, and a hash reference containing the values of the form fields.
If FormMail::process wants to insert any new fields into the mail sent, it
can do so by inserting the name into the array, and the value into the hash.
It can signal the outcome of the processing by its return value: If it
returns a false value, the form will be resent, with the text of the
C<only_if_process_fails> command inserted.

=item C<success_url>

The C<parameter> is the URL that the browser is redirected to if sending the
E-Mail was successful.

=item C<failure_url>

The C<parameter> is the URL that the browser is redirected to if sending the
E-Mail was not successful.

=item C<only_if_first_time>

The contents of C<parameter> is sent as text, if B<FormMail> is called with
the C<GET> request method (which is the usual way it is called before processing).

=item C<only_if_missing_required>

The contents of C<parameter> is sent as text on re-sending the form (that
is, if B<FormMail> is called with the C<POST> request method - which happens
when the form gets submitted), and required input fields have not been
filled in.

=item C<only_if_process_fails>

The contents of C<parameter> is sent as text on re-sending the form (that
is, if B<FormMail> is called with the C<POST> request method - which happens
when the form gets submitted), and the C<process> file returned false.

=back

=head1 CONFIGURATION

Near the top of B<FormMail.pl> is a configuration section with the following
entries:

=over 4

=item C<$mailHost>

The IP address or host name of the mail host to try and send the mail to.

=item C<$webmaster>

A mail address to turn to in case of problems. This should be a mail address
of someone responsible for the site, and is incorporated as a C<LINK
REV="MADE"> tag into the C<HEAD>s of all HTML pages sent.

=item C<$failFile>

A filename to save the mail to, in case the mailing process failed.

=item C<$default_is_required>

What to assume at the top of a form: required (1) or not_required (0).

=item C<$debug>

If this variable is defined, it is taken as a filename to redirect STDERR to
(in append mode).

=item C<$cgi_save>

If this variable is defined, it is taken as a filename to CGI::save the
received form variables to. When debugging interactively, setting the
environment variable C<FORMMAIL_DEBUG> to a true value will cause C<FormMail>
to read the file, and populate parameters passed from it.

=back

=head1 RESTRICTIONS

=over 4

=item *

The variable C<FormURL> must not be used as a name for a form entry field.

=item *

The C<FORM ACTION> must be a C<mailto> URL.

=item *

Each entry field in the form must have a C<NAME> attribute.

=item *

Each C<OPTION> must have a C<VALUE> attribute, i.e. you can't (yet) use the
option's text as the C<OPTION> field's value.

=item *

In the program, all values associated with the same form variable name are
concatenated into a newline separated string. So they'd better belong to the
same context (i.e. to the same C<SELECT> or C<INPUT TYPE=RADIO|CHECKBOX>
field).

=item *

The parse engine for the (not_)required_if expressions is rather simplistic.
So please don't try to abuse it.

=back

=head1 COPYRIGHT

Copyright (c) 1999 Martin VorlE<auml>nder. All rights reserved.
This program is free software; you may redistribute it and/or modify it
under the same terms as Perl itself.

=head1 BUGS

None that I know of... of course, there will be some.
I'd appreciate any feedback.

=head1 CREDITS

This program relies heavily on the following modules, and owes much to their
authors:

=over 4

=item *

CGI by Lincoln D. Stein

=item *

HTML::Parser by Gisle Aas

=item *

various from the libwww library by Gisle Aas and Martijn Koster

=item *

Net::SMTP from the libnet library by Graham Barr

=back

Many thanks also go to the Perl UseNet community and Larry Wall.

=head1 REVISION LOG

  0.9 (17-Jan-1998)
    Initial version released
  1.0 (02-Jan-1999)
    Reworked for presentation at the German Perl Workshop 1.0

=head1 AUTHOR

Martin VorlE<auml>nder E<lt>mv@pdv-systeme.deE<gt>

=cut
